import pytest


def test_create_destroy(ctx_new):
    """Create and destroy a context"""
    # Context will destroy itself on exit
    pass


def test_core_profile_check(ctx):
    ctx.core_profile_check()


def test_detect_framebuffer(ctx):
    # find a randomly active framebuffer.
    # We just test the call for now
    ctx.detect_framebuffer()


def test_extensions(ctx_new):
    # self.assertTrue("GL_ARB_vertex_array_object" in ctx.extensions)
    # self.assertTrue("GL_ARB_transform_feedback2" in ctx.extensions)
    # self.assertTrue("GL_ARB_shader_subroutine" in ctx.extensions)
    ctx = ctx_new
    assert isinstance(ctx.extensions, set)
    assert len(ctx.extensions) > 0


def test_attributes(ctx_new):
    """Ensure enums are present in the context instance"""
    ctx = ctx_new
    # Flags
    assert isinstance(ctx.NOTHING, int)
    assert isinstance(ctx.BLEND, int)
    assert isinstance(ctx.DEPTH_TEST, int)
    assert isinstance(ctx.CULL_FACE, int)
    assert isinstance(ctx.RASTERIZER_DISCARD, int)
    assert isinstance(ctx.PROGRAM_POINT_SIZE, int)

    # Primitive modes
    assert isinstance(ctx.POINTS, int)
    assert isinstance(ctx.LINES, int)
    assert isinstance(ctx.LINE_LOOP, int)
    assert isinstance(ctx.LINE_STRIP, int)
    assert isinstance(ctx.TRIANGLES, int)
    assert isinstance(ctx.TRIANGLE_STRIP, int)
    assert isinstance(ctx.TRIANGLE_FAN, int)
    assert isinstance(ctx.LINES_ADJACENCY, int)
    assert isinstance(ctx.LINE_STRIP_ADJACENCY, int)
    assert isinstance(ctx.TRIANGLES_ADJACENCY, int)
    assert isinstance(ctx.TRIANGLE_STRIP_ADJACENCY, int)
    assert isinstance(ctx.PATCHES, int)

    # Texture filters
    assert isinstance(ctx.LINEAR, int)
    assert isinstance(ctx.NEAREST, int)
    assert isinstance(ctx.NEAREST_MIPMAP_NEAREST, int)
    assert isinstance(ctx.LINEAR_MIPMAP_LINEAR, int)
    assert isinstance(ctx.LINEAR_MIPMAP_NEAREST, int)
    assert isinstance(ctx.NEAREST_MIPMAP_LINEAR, int)

    # Blend functions
    assert isinstance(ctx.ZERO, int)
    assert isinstance(ctx.ONE, int)
    assert isinstance(ctx.SRC_COLOR, int)
    assert isinstance(ctx.ONE_MINUS_SRC_COLOR, int)
    assert isinstance(ctx.SRC_ALPHA, int)
    assert isinstance(ctx.ONE_MINUS_SRC_ALPHA, int)
    assert isinstance(ctx.DST_ALPHA, int)
    assert isinstance(ctx.ONE_MINUS_DST_ALPHA, int)
    assert isinstance(ctx.DST_COLOR, int)
    assert isinstance(ctx.ONE_MINUS_DST_COLOR, int)

    # Blend shortcuts
    assert isinstance(ctx.DEFAULT_BLENDING, tuple)
    assert isinstance(ctx.ADDITIVE_BLENDING, tuple)
    assert isinstance(ctx.PREMULTIPLIED_ALPHA, tuple)

    # Blend equations
    assert isinstance(ctx.FUNC_ADD, int)
    assert isinstance(ctx.FUNC_SUBTRACT, int)
    assert isinstance(ctx.FUNC_REVERSE_SUBTRACT, int)
    assert isinstance(ctx.MIN, int)
    assert isinstance(ctx.MAX, int)

    # Provoking vertex
    assert isinstance(ctx.FIRST_VERTEX_CONVENTION, int)
    assert isinstance(ctx.LAST_VERTEX_CONVENTION, int)

    assert ctx == ctx

    assert ctx.line_width == 1.0
    assert ctx.point_size == 1.0
    assert isinstance(ctx.max_integer_samples, int)
    assert isinstance(ctx.max_texture_units, int)
    assert ctx.default_texture_unit == ctx.max_texture_units - 1
    assert ctx.screen is None
    assert ctx.wireframe is False
    assert ctx.front_face == "ccw"
    assert ctx.cull_face == "front_and_back"
    assert ctx.patch_vertices == 3

    ctx.enable(ctx.NOTHING)
    ctx.disable(ctx.NOTHING)
    ctx.finish()


def test_enable_direct(ctx_new):
    # We already support this, but it's a safe value
    ctx = ctx_new
    GL_PROGRAM_POINT_SIZE = 0x8642

    ctx.enable_direct(GL_PROGRAM_POINT_SIZE)
    assert ctx.error == "GL_NO_ERROR"

    ctx.disable_direct(GL_PROGRAM_POINT_SIZE)
    assert ctx.error == "GL_NO_ERROR"

def test_depth_clamp_range(ctx):
    ctx.depth_clamp_range = 0, 1
    assert ctx.error == "GL_NO_ERROR"

    ctx.depth_clamp_range = None
    assert ctx.error == "GL_NO_ERROR"

def test_info(ctx):
    assert isinstance(ctx.info, dict)
    assert len(ctx.info) > 50
    assert ctx.info["GL_MAX_GEOMETRY_OUTPUT_VERTICES"] >= 256


def test_polygon_offset(ctx):
    ctx.polygon_offset = 0.0, 1.0
    assert ctx.polygon_offset == (0.0, 1.0)
    ctx.polygon_offset = 1.0, 0.0
    assert ctx.polygon_offset == (1.0, 0.0)
    ctx.polygon_offset = -1.0, -1.0
    assert ctx.polygon_offset == (-1.0, -1.0)


def test_context_gc_modes(ctx_new):
    """Check gc mode default and set different modes"""
    ctx = ctx_new
    # gc mode should be None by default
    assert ctx.gc_mode is None
    ctx.gc_mode = None
    ctx.gc_mode = "context_gc"
    ctx.gc_mode = "auto"
    with pytest.raises(ValueError):
        ctx.gc_mode = "something"


def test_context_gc(ctx_new):
    """Simple usage of context_gc"""
    ctx = ctx_new
    ctx.gc_mode = "context_gc"

    # Buffer
    buff = ctx.buffer(reserve=1024)
    buff = None
    assert ctx.gc() == 1

    # Texture
    tex = ctx.texture((10, 10), 4)
    tex = None
    assert ctx.gc() == 1

    # Texture Array
    tex_array = ctx.texture_array((10, 10, 10), 4)
    tex_array = None
    assert ctx.gc() == 1

    # Texture cube
    tex_cube = ctx.texture_cube((10, 10), 4)
    tex_cube = None
    assert ctx.gc() == 1

    # Texture3D
    tex_3d = ctx.texture3d((10, 10, 10), 4)
    tex_3d = None
    assert ctx.gc() == 1

    # Sampler
    sampler = ctx.sampler()
    sampler = None
    assert ctx.gc() == 1

    # Renderbuffer
    rb = ctx.renderbuffer((10, 19))
    rb = None
    assert ctx.gc() == 1

    # Program
    prog = ctx.program(
        vertex_shader="""
        #version 330
        out float value;
        void main() {
            value = float(gl_VertexID);
        }
        """
    )
    prog = None
    assert ctx.gc() == 1

    # Framebuffer
    fbo = ctx.framebuffer(
        color_attachments=[ctx.texture((10, 10), 4)],
        depth_attachment=ctx.depth_texture((10, 10))
    )
    fbo = None
    assert ctx.gc() == 3

    # # Compute Shader
    # cs = ctx.compute_shader(
    #     """
    #     #version 450
        
    #     layout(local_size_x=4, local_size_y=4) in;

    #     layout(rgba8, binding=0) uniform imageCube img_in;
    #     layout(rgba8, binding=1) uniform imageCube img_out;

    #     void main() {
    #         for (int i = 0; i < 6; i++) {
    #             vec4 fragment = imageLoad(img_in, ivec3(gl_LocalInvocationID.xy, i));
    #             imageStore(img_out, ivec3(gl_LocalInvocationID.xy, i), fragment);
    #         }
    #     }
    #     """
    # )
    # cs = None
    # self.assertEqual(ctx.gc(), 1)

# # --- Too complex to test right now

# # def test_context_switch(self):
# #     """Ensure context switching is working"""
# #     ctx1 = moderngl.create_context(standalone=True)
# #     ctx2 = moderngl.create_context(standalone=True)

# #     with ctx1 as ctx:
# #         buffer1 = ctx.buffer(reserve=1024)

# #     with ctx2 as ctx:
# #         buffer2 = ctx.buffer(reserve=1024)

# #     self.assertEqual(buffer1.glo, buffer2.glo)
# #     ctx1.release()
# #     ctx2.release()

# # def test_exit(self):
# #     """Ensure the previous context was activated on exit"""
# #     ctx1 = moderngl.create_context(standalone=True)
# #     ctx2 = moderngl.create_context(standalone=True)

# #     with ctx1 as ctx:
# #         ctx.buffer(reserve=1024)

# #     # Will error out if no context is active "moderngl.Error: cannot create buffer"
# #     ctx1.buffer(reserve=1024)

# #     ctx1.release()
# #     ctx2.release()

# # def test_share(self):
# #     """Create resources with shared context"""
# #     if platform.system().lower() in ["darwin", "linux"]:
# #         self.skipTest('Context sharing not supported on darwin')

# #     data1 = numpy.array([1, 2, 3, 4], dtype='u1')
# #     data2 = numpy.array([4, 3, 2, 1], dtype='u1')

# #     ctx1 = moderngl.create_context(standalone=True)
# #     ctx2 = moderngl.create_context(standalone=True, share=True)

# #     with ctx1 as ctx:
# #         b1 = ctx.buffer(data=data1)

# #     with ctx2 as ctx:
# #         b2 = ctx.buffer(data=data2)

# #     # Because the resources are shared the name should increment
# #     self.assertEqual(b1.glo, 1)
# #     self.assertEqual(b2.glo, 2)

# #     # Ensure we can read the same buffer data in both contexts
# #     with ctx1:
# #         self.assertEqual(b1.read(), b'\x01\x02\x03\x04')
# #         self.assertEqual(b2.read(), b'\x04\x03\x02\x01')

# #     with ctx2:
# #         self.assertEqual(b1.read(), b'\x01\x02\x03\x04')
# #         self.assertEqual(b2.read(), b'\x04\x03\x02\x01')

# #     ctx1.release()
# #     ctx2.release()




def test_1(ctx):
    fbo = ctx.framebuffer(ctx.renderbuffer((4, 4)))

    fbo.clear(0.0, 0.0, 0.0, 0.0)

    assert fbo.read(components=4) == b'\x00\x00\x00\x00' * 16

    fbo.color_mask = (True, False, True, False)
    fbo.clear(0x19 / 255, 0x33 / 255, 0x4c / 255, 0x66 / 255)

    assert fbo.read(components=4) == b'\x19\x00\x4c\x00' * 16

    fbo.color_mask = (False, True, False, True)
    fbo.clear(0x7f / 255, 0x99 / 255, 0xb2 / 255, 0xcc / 255)

    assert fbo.read(components=4) == b'\x19\x99\x4c\xcc' * 16


def test_2(ctx):
    fbo = ctx.framebuffer([
        ctx.renderbuffer((4, 4)),
        ctx.renderbuffer((4, 4)),
    ])

    fbo.clear(1.0, 1.0, 1.0, 1.0)

    assert fbo.read(components=4, attachment=0) == b'\xff\xff\xff\xff' * 16
    assert fbo.read(components=4, attachment=1) == b'\xff\xff\xff\xff' * 16

    fbo.color_mask = (
        (True, False, True, False),
        (False, True, False, True),
    )

    fbo.clear(0x19 / 255, 0x33 / 255, 0x4c / 255, 0x66 / 255)

    assert fbo.read(components=4, attachment=0) == b'\x19\xff\x4c\xff' * 16
    assert fbo.read(components=4, attachment=1) == b'\xff\x33\xff\x66' * 16

    fbo.color_mask = (
        (False, True, False, True),
        (True, False, True, False),
    )

    fbo.clear(0x7f / 255, 0x99 / 255, 0xb2 / 255, 0xcc / 255)

    assert fbo.read(components=4, attachment=0) == b'\x19\x99\x4c\xcc' * 16
    assert fbo.read(components=4, attachment=1) == b'\x7f\x33\xb2\x66' * 16



import struct
import pytest
import moderngl


def test_vao_attribs(ctx):

    program = ctx.program(
        vertex_shader='''
            #version 330

            layout(location = 0) in vec4 in_vert;
            out vec4 out_vert;

            void main() {
                out_vert = in_vert;
            }
        ''',
        varyings=['out_vert']
    )

    vbo = ctx.buffer(struct.pack('12f', 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12))
    res = ctx.buffer(reserve=struct.calcsize('8f'))

    vao = ctx.simple_vertex_array(program, vbo, 'in_vert')
    vao.transform(res, moderngl.POINTS, 2)

    for a, b in zip(struct.unpack('8f', res.read()), (1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0)):
        assert pytest.approx(a) == b

    vao.bind(0, 'f', vbo, '4f', offset=4, stride=0, divisor=0)
    vao.transform(res, moderngl.POINTS, 2)

    for a, b in zip(struct.unpack('8f', res.read()), (2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0)):
        assert pytest.approx(a) == b

    vao.bind(0, 'f', vbo, '4f', offset=8, stride=20, divisor=0)
    vao.transform(res, moderngl.POINTS, 2)

    for a, b in zip(struct.unpack('8f', res.read()), (3.0, 4.0, 5.0, 6.0, 8.0, 9.0, 10.0, 11.0)):
        assert pytest.approx(a) == b

    vao.bind(0, 'f', vbo, '4f', offset=12, stride=0, divisor=1)
    vao.transform(res, moderngl.POINTS, 2)

    for a, b in zip(struct.unpack('8f', res.read()), (4.0, 5.0, 6.0, 7.0, 4.0, 5.0, 6.0, 7.0)):
        assert pytest.approx(a) == b



import struct
from array import array
import pytest
import moderngl


@pytest.fixture(scope='module', autouse=True)
def vao(ctx_static):
    prog = ctx_static.program(
        vertex_shader='''
            #version 330
            in vec2 vert;
            out vec2 text;
            void main() {
                gl_Position = vec4(vert * 2.0 - 1.0, 0.0, 1.0);
                text = vert;
            }
        ''',
        fragment_shader='''
            #version 330
            uniform sampler2D Texture;
            in vec2 text;
            out vec4 color;
            void main() {
                color = texture(Texture, text);
            }
        ''',
    )
    vbo = ctx_static.buffer(struct.pack('8f', 0, 0, 0, 1, 1, 0, 1, 1))
    return ctx_static.simple_vertex_array(prog, vbo, 'vert')


def test_texture_properties(ctx):
    tex = ctx.texture((16, 16), 4)
    assert tex.width == 16
    assert tex.height == 16
    assert tex.size == (16, 16)
    assert tex.components == 4
    assert tex.depth is False
    assert tex.samples == 0
    assert tex.glo > 0
    assert tex.filter == (moderngl.LINEAR, moderngl.LINEAR)
    assert tex.repeat_x is True
    assert tex.repeat_y is True
    assert tex.dtype == 'f1'
    assert tex.anisotropy == 0.0

    tex.anisotropy = ctx.max_anisotropy
    assert tex.anisotropy == ctx.max_anisotropy
    tex.build_mipmaps()
    assert tex.filter == (moderngl.LINEAR_MIPMAP_LINEAR, moderngl.LINEAR)


def test_1(ctx, vao):
    assert vao.mode == moderngl.TRIANGLES
    fbo = ctx.framebuffer(ctx.renderbuffer((16, 16)))
    pixels = struct.pack('16B', 255, 0, 0, 255, 0, 255, 0, 255, 0, 0, 255, 255, 0, 0, 0, 255)
    texture = ctx.texture((2, 2), 4, pixels)

    fbo.use()
    texture.use()

    texture.filter = (moderngl.LINEAR, moderngl.LINEAR)
    vao.render(moderngl.TRIANGLE_STRIP)

    r, g, b = struct.unpack('3B', fbo.read((8, 8, 1, 1)))
    assert (abs(r - 49) < 8 and abs(g - 63) < 8 and abs(b - 63) < 8)

    texture.filter = (moderngl.NEAREST, moderngl.NEAREST)
    vao.render(moderngl.TRIANGLE_STRIP)

    r, g, b = struct.unpack('3B', fbo.read((6, 6, 1, 1)))
    assert (abs(r - 255) < 8 and abs(g - 0) < 8 and abs(b - 0) < 8)

    r, g, b = struct.unpack('3B', fbo.read((9, 6, 1, 1)))
    assert (abs(r - 0) < 8 and abs(g - 255) < 8 and abs(b - 0) < 8)

    r, g, b = struct.unpack('3B', fbo.read((6, 9, 1, 1)))
    assert (abs(r - 0) < 8 and abs(g - 0) < 8 and abs(b - 255) < 8)

    r, g, b = struct.unpack('3B', fbo.read((9, 9, 1, 1)))
    assert (abs(r - 0) < 8 and abs(g - 0) < 8 and abs(b - 0) < 8)


def test_2(ctx, vao):
    fbo = ctx.framebuffer(ctx.renderbuffer((16, 16)))
    pixels1 = struct.pack('16B', 255, 0, 0, 255, 0, 255, 0, 255, 255, 0, 0, 255, 0, 255, 0, 255)
    pixels2 = struct.pack('16B', 20, 30, 40, 255, 80, 70, 60, 255, 20, 30, 40, 255, 80, 70, 60, 255)
    texture = ctx.texture((2, 2), 4, pixels1)

    fbo.use()
    texture.use()

    texture.filter = (moderngl.NEAREST, moderngl.NEAREST)
    vao.render(moderngl.TRIANGLE_STRIP)

    r, g, b = struct.unpack('3B', fbo.read((5, 7, 1, 1)))
    assert (abs(r - 255) < 8 and abs(g - 0) < 8 and abs(b - 0) < 8)

    r, g, b = struct.unpack('3B', fbo.read((10, 7, 1, 1)))
    assert (abs(r - 0) < 8 and abs(g - 255) < 8 and abs(b - 0) < 8)

    texture.write(pixels2)

    texture.filter = (moderngl.NEAREST, moderngl.NEAREST)
    vao.render(moderngl.TRIANGLE_STRIP)

    r, g, b = struct.unpack('3B', fbo.read((5, 7, 1, 1)))
    assert (abs(r - 20) < 8 and abs(g - 30) < 8 and abs(b - 40) < 8)

    r, g, b = struct.unpack('3B', fbo.read((10, 7, 1, 1)))
    assert (abs(r - 80) < 8 and abs(g - 70) < 8 and abs(b - 60) < 8)


def test_3(ctx, vao):
    fbo = ctx.framebuffer(ctx.renderbuffer((16, 16)))
    pixels1 = struct.pack('16B', 255, 0, 0, 255, 0, 255, 0, 255, 255, 0, 0, 255, 0, 255, 0, 255)
    pixels2 = struct.pack('16B', 20, 30, 40, 255, 80, 70, 60, 255, 20, 30, 40, 255, 80, 70, 60, 255)
    texture = ctx.texture((2, 2), 4)

    pbo1 = ctx.buffer(pixels1)
    pbo2 = ctx.buffer(pixels2)

    fbo.use()
    texture.use()

    texture.write(pbo1)
    texture.filter = (moderngl.NEAREST, moderngl.NEAREST)
    vao.render(moderngl.TRIANGLE_STRIP)

    r, g, b = struct.unpack('3B', fbo.read((5, 7, 1, 1)))
    assert (abs(r - 255) < 8 and abs(g - 0) < 8 and abs(b - 0) < 8)

    r, g, b = struct.unpack('3B', fbo.read((10, 7, 1, 1)))
    assert (abs(r - 0) < 8 and abs(g - 255) < 8 and abs(b - 0) < 8)

    texture.write(pbo2)
    texture.filter = (moderngl.NEAREST, moderngl.NEAREST)
    vao.render(moderngl.TRIANGLE_STRIP)

    r, g, b = struct.unpack('3B', fbo.read((5, 7, 1, 1)))
    assert (abs(r - 20) < 8 and abs(g - 30) < 8 and abs(b - 40) < 8)

    r, g, b = struct.unpack('3B', fbo.read((10, 7, 1, 1)))
    assert (abs(r - 80) < 8 and abs(g - 70) < 8 and abs(b - 60) < 8)


def test_4(ctx):
    pixels = struct.pack('16B', 255, 0, 0, 255, 0, 255, 0, 255, 255, 0, 0, 255, 0, 255, 0, 255)
    texture = ctx.texture((2, 2), 4, pixels)
    assert texture.read() == pixels


def test_5(ctx):
    pixels = struct.pack('16B', 255, 0, 0, 255, 0, 255, 0, 255, 255, 0, 0, 255, 0, 255, 0, 255)
    texture = ctx.texture((2, 2), 4, pixels)
    buf = bytearray(len(pixels))
    texture.read_into(buf)

    assert bytes(buf) == pixels


def test_6(ctx):
    pixels = struct.pack('16B', 255, 0, 0, 255, 0, 255, 0, 255, 255, 0, 0, 255, 0, 255, 0, 255)
    texture = ctx.texture((2, 2), 4, pixels)
    pbo = ctx.buffer(reserve=len(pixels))
    texture.read_into(pbo)

    assert pbo.read() == pixels


def test_override_internalformat(ctx):
    """Ensure no errors occur when overriding internalformat"""
    if "GL_EXT_texture_sRGB" not in ctx.extensions:
        pytest.skip('GL_EXT_texture_sRGB extension not supported')

    GL_SRGB8 = 0x8C41
    pixels = struct.pack('16B', 255, 0, 0, 255, 0, 255, 0, 255, 255, 0, 0, 255, 0, 255, 0, 255)
    texture = ctx.texture((2, 2), 4, pixels, internal_format=GL_SRGB8)
    _ = texture.read()
    assert ctx.error == "GL_NO_ERROR"


def test_normalized_textures(ctx, vao):
    """8 and 16 bit normalized integer textures"""
    ni1 = ctx.texture((4, 4), 4, dtype="ni1")
    ni2 = ctx.texture((4, 4), 4, dtype="ni2")
    nu1 = ctx.texture((4, 4), 4, dtype="nu1")
    nu2 = ctx.texture((4, 4), 4, dtype="nu2")

    nu1_data = array('B', [255] * 16 * 4).tobytes()
    nu2_data = array('H', [65535] * 16 * 4).tobytes()
    ni1_data = array('b', [127] * 16 * 4).tobytes()
    ni2_data = array('h', [32767] * 16 * 4).tobytes()

    nu1.write(nu1_data)
    nu2.write(nu2_data)
    ni1.write(ni1_data)
    ni2.write(ni2_data)

    assert nu1.read() == nu1_data
    assert nu2.read() == nu2_data
    assert ni1.read() == ni1_data
    assert ni2.read() == ni2_data

    fbo = ctx.simple_framebuffer((4, 4))
    fbo.use()

    # Render these textures to an RGBA8 framebuffer and ensure the result is a pure white color (FF)
    fbo.clear()
    nu1.use()
    vao.render()
    assert fbo.read(viewport=(0, 0, 1, 1), components=4, dtype="f1") == b'\xff\xff\xff\xff'

    fbo.clear()
    nu2.use()
    vao.render()
    assert fbo.read(viewport=(0, 0, 1, 1), components=4, dtype="f1") == b'\xff\xff\xff\xff'

    fbo.clear()
    ni1.use()
    vao.render()
    assert fbo.read(viewport=(0, 0, 1, 1), components=4, dtype="f1") == b'\xff\xff\xff\xff'

    fbo.clear()
    ni2.use()
    vao.render()
    assert fbo.read(viewport=(0, 0, 1, 1), components=4, dtype="f1") == b'\xff\xff\xff\xff'


def test_depth_texture_write(ctx):
    """Write data into depth texture"""
    size = 4, 4
    fbo = ctx.framebuffer(
        color_attachments=[ctx.texture(size, 4)],
        depth_attachment=ctx.depth_texture(size),
    )
    fbo.clear(depth=1.0)

    first = struct.unpack("16f", fbo.depth_attachment.read())
    second = (1.0,) * 16
    for a, b in zip(first, second):
        assert pytest.approx(a, abs=1.0e-6) == b

    fbo.depth_attachment.write(struct.pack("16f", *([0.5] * 16)))
    first = struct.unpack("16f", fbo.depth_attachment.read())
    second = (0.5,) * 16
    for a, b in zip(first, second):
        assert pytest.approx(a, abs=1.0e-6) == b



import pytest


def test_1(ctx):
    buf = ctx.buffer(b'abc')

    with pytest.raises(Exception):
        buf.write(b'1234')

    with pytest.raises(Exception):
        buf.write(b'u', offset=-1)

    with pytest.raises(Exception):
        buf.write(b'abc', offset=1)


def test_2(ctx):
    buf = ctx.buffer(b'123456789')

    with pytest.raises(Exception):
        buf.write_chunks(b'12345', 0, 1, 4)

    with pytest.raises(Exception):
        buf.write_chunks(b'12345', 0, 1, 6)

    with pytest.raises(Exception):
        buf.write_chunks(b'yyyyyn', 4, 1, 6)

    with pytest.raises(Exception):
        buf.write_chunks(b'yyyyyn', 4, -1, 6)

    with pytest.raises(Exception):
        buf.write_chunks(b'yn', -1, 2, 1)


def test_3(ctx):
    buf = ctx.buffer(b'123456789')

    with pytest.raises(Exception):
        buf.write_chunks(b'yyyyyn', 2, 3, 3)

    with pytest.raises(Exception):
        buf.write_chunks(b'ynyyyy', -1, -3, 3)

    with pytest.raises(Exception):
        buf.write_chunks(b'yyyyny', -4, -3, 3)


def test_4(ctx):
    buf = ctx.buffer(b'123456789')

    with pytest.raises(Exception):
        buf.write_chunks(b'yyynyy', 0, 2, 2)



import moderngl


def test_1(ctx):
    prog = ctx.program(
        vertex_shader='''
            #version 330

            in vec2 in_vert;
            out vec2 v_vert;

            uniform mat2 Mvp;
            uniform vec2 Position;
            uniform float Scale;

            void main() {
                v_vert = Position + in_vert * Scale;
                gl_Position = vec4(Mvp * v_vert, 0.0, 1.0);
            }
        ''',
        fragment_shader='''
            #version 330

            uniform bool UseTexture;
            uniform sampler2D Texture;

            uniform vec3 Color;
            uniform float Alpha;

            in vec2 v_vert;
            out vec4 f_color;

            void main() {
                if (UseTexture) {
                    f_color = texture(Texture, v_vert);
                    f_color.rgb *= Color;
                    f_color.a *= Alpha;
                } else {
                    f_color = vec4(Color, Alpha);
                }
            }
        ''',
    )

    assert 'in_vert' in prog
    assert 'Position' in prog
    assert 'Scale' in prog
    assert 'Mvp' in prog
    assert 'UseTexture' in prog
    assert 'Texture' in prog
    assert 'Color' in prog
    assert 'Alpha' in prog

    assert isinstance(prog['in_vert'], moderngl.Attribute)
    assert isinstance(prog['Position'], moderngl.Uniform)
    assert isinstance(prog['Scale'], moderngl.Uniform)
    assert isinstance(prog['Mvp'], moderngl.Uniform)
    assert isinstance(prog['UseTexture'], moderngl.Uniform)
    assert isinstance(prog['Texture'], moderngl.Uniform)
    assert isinstance(prog['Color'], moderngl.Uniform)
    assert isinstance(prog['Alpha'], moderngl.Uniform)



import os
import re


def read_file(name):
    f = open(name)
    res = f.read()
    f.close()
    return res


here = os.path.dirname(os.path.abspath(__file__))
repo = os.path.dirname(here)


# def test_version():
#     setup = read_file(os.path.join(repo, 'setup.py'))
#     init = read_file(os.path.join(repo, 'moderngl', '__init__.py'))
#     docs = read_file(os.path.join(repo, 'docs', 'conf.py'))

#     match0 = re.search(r'version=\'(\d+\.\d+\.\d+)\'', setup, flags=re.M)

#     match1 = re.search(r'__version__ = \'(\d+\.\d+\.\d+)\'', init, flags=re.M)
#     match2 = re.search(r'version = \'(\d+\.\d+\.\d+)\'', docs, flags=re.M)
#     match3 = re.search(r'release = \'(\d+\.\d+\.\d+)\'', docs, flags=re.M)

#     assert match0.group(1) == match1.group(1), 'Version error: moderngl/__init__.py'
#     assert match0.group(1) == match2.group(1), 'Version error: docs/conf.py'
#     assert match0.group(1) == match3.group(1), 'Version error: docs/conf.py'



import pytest


def test_renderbuffer(ctx):
    rbo = ctx.renderbuffer((4, 8))
    assert rbo.size == (4, 8)
    assert rbo.width == 4
    assert rbo.height == 8
    assert rbo.samples == 0
    assert rbo.depth is False
    assert rbo.components == 4
    assert rbo.dtype == 'f1'
    assert rbo == rbo


def test_multisample_renderbuffer(ctx):
    if ctx.max_samples < 2:
        pytest.skip('multisampling is not supported')

    rbo = ctx.renderbuffer((4, 4), samples=2)
    assert rbo.size == (4, 4)
    assert rbo.samples == 2
    assert rbo.depth is False


def test_depth_renderbuffer(ctx):
    rbo = ctx.depth_renderbuffer((4, 4))
    assert rbo.size == (4, 4)
    assert rbo.samples == 0
    assert rbo.depth is True


def test_multisample_depth_renderbuffer(ctx):
    if ctx.max_samples < 2:
        pytest.skip('multisampling is not supported')

    rbo = ctx.depth_renderbuffer((4, 4), samples=2)
    assert rbo.size == (4, 4)
    assert rbo.samples == 2
    assert rbo.depth is True


def test_renderbuffer_invalid_samples(ctx):
    if ctx.max_samples < 2:
        pytest.skip('multisampling is not supported')

    with pytest.raises(Exception, match='samples is invalid'):
        ctx.renderbuffer((4, 4), samples=3)



from array import array
import struct
import pytest


def test_1(ctx):
    if ctx.version_code < 430:
        pytest.skip('compute shaders not supported')

    compute_shader = ctx.compute_shader('''
        #version 430

        layout (local_size_x = 1, local_size_y = 1) in;

        layout (std430, binding = 1) buffer Input {
            float v1[4];
        };

        layout (std430, binding = 2) buffer Output {
            float v2[4];
        };

        void main() {
            v2[0] = v1[3];
            v2[1] = v1[2];
            v2[2] = v1[1];
            v2[3] = v1[0];
        }
    ''')

    buf1 = ctx.buffer(struct.pack('4f', 1.0, 2.0, 3.0, 4.0))
    buf2 = ctx.buffer(struct.pack('4f', 0.0, 0.0, 0.0, 0.0))

    buf1.bind_to_storage_buffer(1)
    buf2.bind_to_storage_buffer(2)

    compute_shader.run()

    a, b, c, d = struct.unpack('4f', buf2.read())

    assert pytest.approx(a) == 4.0
    assert pytest.approx(b) == 3.0
    assert pytest.approx(c) == 2.0
    assert pytest.approx(d) == 1.0


def test_image(ctx):
    if ctx.version_code < 430:
        pytest.skip('compute shaders not supported')

    texture = ctx.texture((100, 100), 4)
    texture.bind_to_image(0, read=True, write=True)
    assert ctx.error == 'GL_NO_ERROR'
    texture.release()


def test_image_float(ctx):
    if ctx.version_code < 430:
        pytest.skip('compute shaders not supported')

    texture = ctx.texture((100, 100), 4, dtype='f4')
    texture.bind_to_image(0, read=True, write=True)
    assert ctx.error == 'GL_NO_ERROR'
    texture.release()


def test_image_wrong_format(ctx):
    if ctx.version_code < 430:
        pytest.skip('compute shaders not supported')

    texture = ctx.texture((100, 100), 4)
    texture.bind_to_image(0, read=True, write=True, format=13371337)
    assert ctx.error == 'GL_INVALID_VALUE'
    texture.release()


def test_3d_image(ctx):
    if ctx.version_code < 430:
        pytest.skip('compute shaders not supported')

    program = ctx.compute_shader(
        """
        #version 430

        layout(local_size_x=4, local_size_y=4, local_size_z=4) in;

        layout(rgba32f, binding=0) uniform image3D img_in;
        layout(rgba32f, binding=1) uniform image3D img_out;

        void main() {
            vec4 fragment = imageLoad(img_in, ivec3(gl_LocalInvocationID.xyz));
            imageStore(img_out, ivec3(gl_LocalInvocationID.xyz), fragment);
        }
        """
    )
    tex_in = ctx.texture3d((4, 4, 4), 4, data=array('f', [v for v in range(4 * 4 * 4 * 4)]), dtype="f4")
    tex_out = ctx.texture3d((4, 4, 4), 4, dtype="f4")

    tex_in.bind_to_image(0, read=True, write=False)
    tex_out.bind_to_image(1, read=False, write=True)
    program.run(group_x=1)
    assert ctx.error == 'GL_NO_ERROR'

    data_in = struct.unpack("256f", tex_in.read())
    data_out = struct.unpack("256f", tex_out.read())

    assert data_in == data_out


def test_texture_array_image(ctx):
    if ctx.version_code < 430:
        pytest.skip('compute shaders not supported')

    program = ctx.compute_shader(
        """
        #version 430

        layout(local_size_x=4, local_size_y=4, local_size_z=4) in;

        layout(rgba32f, binding=0) uniform image2DArray img_in;
        layout(rgba32f, binding=1) uniform image2DArray img_out;

        void main() {
            vec4 fragment = imageLoad(img_in, ivec3(gl_LocalInvocationID.xyz));
            imageStore(img_out, ivec3(gl_LocalInvocationID.xyz), fragment);
        }
        """
    )
    tex_in = ctx.texture_array((4, 4, 4), 4, data=array('f', [v for v in range(4 * 4 * 4 * 4)]), dtype="f4")
    tex_out = ctx.texture_array((4, 4, 4), 4, dtype="f4")

    tex_in.bind_to_image(0, read=True, write=False)
    tex_out.bind_to_image(1, read=False, write=True)
    program.run(group_x=1)

    data_in = struct.unpack("256f", tex_in.read())
    data_out = struct.unpack("256f", tex_out.read())

    assert data_in == data_out


def test_texture_cube_image(ctx):
    if ctx.version_code < 430:
        pytest.skip('compute shaders not supported')

    program = ctx.compute_shader(
        """
        #version 450

        layout(local_size_x=4, local_size_y=4) in;

        layout(rgba8, binding=0) uniform imageCube img_in;
        layout(rgba8, binding=1) uniform imageCube img_out;

        void main() {
            for (int i = 0; i < 6; i++) {
                vec4 fragment = imageLoad(img_in, ivec3(gl_LocalInvocationID.xy, i));
                imageStore(img_out, ivec3(gl_LocalInvocationID.xy, i), fragment);
            }
        }
        """
    )
    data = ([1] * 64) + ([2] * 64) + ([3] * 64) + ([4] * 64) + ([5] * 64) + ([6] * 64)
    tex_in = ctx.texture_cube((4, 4), 4, data=array('B', data), dtype="f1")
    tex_out = ctx.texture_cube((4, 4), 4, data=array('B', [0] * 64 * 6), dtype="f1")
    tex_in.bind_to_image(0, read=True, write=False)
    tex_out.bind_to_image(1, read=False, write=True)
    program.run(group_x=1)

    for face in range(0, 6):
        assert struct.unpack("64B", tex_in.read(face)) == struct.unpack("64B", tex_out.read(face))


def test_ssbo_binding(ctx):
    if ctx.version_code < 430:
        pytest.skip('compute shaders not supported')

    program = ctx.compute_shader(
        """
        #version 430

        layout(local_size_x=1, local_size_y=1, local_size_z=1) in;

        layout(std430, binding=0) buffer ssbo_in {
            float data[4];
        } in_data;

        layout(std430, binding=1) buffer ssbo_out {
            float data[4];
        } out_data;

        void main() {
            out_data.data[gl_LocalInvocationIndex] = in_data.data[gl_LocalInvocationIndex];
        }
        """
    )
    ssbo_in = program['ssbo_in']
    ssbo_out = program['ssbo_out']

    assert ssbo_in.name == 'ssbo_in'
    assert ssbo_out.name == 'ssbo_out'

    assert ssbo_in.binding == 0
    assert ssbo_in.value == 0
    assert ssbo_out.binding == 1
    assert ssbo_out.value == 1

    ssbo_in.binding = 2
    ssbo_out.binding = 3
    assert ssbo_in.binding == 2
    assert ssbo_in.value == 2
    assert ssbo_out.binding == 3
    assert ssbo_out.value == 3



from array import array
import pytest
import struct


@pytest.fixture(scope='module', autouse=True)
def res(ctx_static):
    """Global buffer we transform results into"""
    return ctx_static.buffer(reserve=1024)


def test_float_uniform(ctx, res):
    prog = ctx.program(
        vertex_shader='''
        #version 330
        uniform float Uniform;
        in float v_in;
        out float v_out;
        void main() {
            v_out = Uniform * v_in;
        }
    ''',
        varyings=['v_out']
    )
    vbo = ctx.buffer(struct.pack('f', 1.0))
    vao = ctx.simple_vertex_array(prog, vbo, 'v_in')

    prog['Uniform'].value = 7.0
    vao.transform(res)

    val = struct.unpack('f', res.read(4))[0]
    assert pytest.approx(val) == 7.0


def test_int_uniform(ctx, res):
    prog = ctx.program(
        vertex_shader='''
            #version 330
            uniform int Uniform;
            in int v_in;
            out int v_out;
            void main() {
                v_out = Uniform * v_in;
            }
        ''',
        varyings=['v_out']
    )
    vbo = ctx.buffer(struct.pack('i', 1))
    vao = ctx.simple_vertex_array(prog, vbo, 'v_in')

    prog['Uniform'].value = -2
    vao.transform(res)

    val = struct.unpack('i', res.read(4))[0]
    assert pytest.approx(val) == -2


def test_vec_uniform(ctx, res):
    prog = ctx.program(
        vertex_shader='''
            #version 330
            uniform vec3 Uniform;
            in vec3 v_in;
            out vec3 v_out;
            void main() {
                v_out = Uniform * v_in;
            }
        ''',
        varyings=['v_out']
    )

    vbo = ctx.buffer(struct.pack('3f', 1.0, 1.0, 1.0))
    vao = ctx.simple_vertex_array(prog, vbo, 'v_in')

    prog['Uniform'].value = (0.5, 1.0, 1.5)
    vao.transform(res)

    x, y, z = struct.unpack('3f', res.read(12))
    assert pytest.approx(x) == 0.5
    assert pytest.approx(y) == 1.0
    assert pytest.approx(z) == 1.5


def test_mat_uniform(ctx, res):
    prog = ctx.program(
        vertex_shader='''
            #version 330
            uniform mat2x3 Uniform;
            in float v_in;
            out mat2x3 v_out;
            void main() {
                v_out = Uniform * v_in;
            }
        ''',
        varyings=['v_out']
    )

    vbo = ctx.buffer(struct.pack('f', 1.0))
    vao = ctx.simple_vertex_array(prog, vbo, 'v_in')

    prog['Uniform'].value = (0.0, 1.0, 2.0, 3.0, 4.0, 5.0)
    vao.transform(res)

    m = struct.unpack('6f', res.read(24))
    assert pytest.approx(m[0]) == 0.0
    assert pytest.approx(m[1]) == 1.0
    assert pytest.approx(m[2]) == 2.0
    assert pytest.approx(m[3]) == 3.0
    assert pytest.approx(m[4]) == 4.0
    assert pytest.approx(m[5]) == 5.0


def test_sampler_2d(ctx):
    """RGBA8 2d sampler"""
    prog = ctx.program(
        vertex_shader="""
        #version 330
        uniform sampler2D tex;
        out float color;
        void main() {
            color = texelFetch(tex, ivec2(gl_VertexID, 0), 0).r;
        }
        """,
        varyings=['color'],
    )
    tex = ctx.texture((4, 1), 1, data=array('B', [127, 0, 255, 64]), dtype='f1')
    buff = ctx.buffer(reserve=4 * 4)
    vao = ctx.vertex_array(prog, [])
    tex.use(0)
    vao.transform(buff, vertices=4)
    data = struct.unpack('4f', buff.read())
    assert pytest.approx(data[0], abs=1.0e-3) == 0.498
    assert pytest.approx(data[1]) == 0.0
    assert pytest.approx(data[2]) == 1.0
    assert pytest.approx(data[3], abs=1.0e-3) == 0.25


def test_sampler_2d_int(ctx):
    prog = ctx.program(
        vertex_shader="""
        #version 330
        uniform isampler2D tex;
        out int color;
        void main() {
            color = texelFetch(tex, ivec2(gl_VertexID, 0), 0).r;
        }
        """,
        varyings=['color'],
    )
    tex = ctx.texture((4, 1), 1, data=array('i', [-1, 0, 1000, 4353454]), dtype='i4')
    buff = ctx.buffer(reserve=4 * 4)
    vao = ctx.vertex_array(prog, [])
    tex.use(0)
    vao.transform(buff, vertices=4)
    data = struct.unpack('4i', buff.read())
    assert data[0] == -1
    assert data[1] == 0
    assert data[2] == 1000
    assert data[3] == 4353454


def test_sampler_2d_uint(ctx):
    prog = ctx.program(
        vertex_shader="""
        #version 330
        uniform usampler2D tex;
        out uint color;
        void main() {
            color = texelFetch(tex, ivec2(gl_VertexID, 0), 0).r;
        }
        """,
        varyings=['color'],
    )
    tex = ctx.texture((4, 1), 1, data=array('I', [0, 500, 1000, 4353454]), dtype='u4')
    buff = ctx.buffer(reserve=4 * 4)
    vao = ctx.vertex_array(prog, [])
    tex.use(0)
    vao.transform(buff, vertices=4)
    data = struct.unpack('4I', buff.read())
    assert data[0] == 0
    assert data[1] == 500
    assert data[2] == 1000
    assert data[3] == 4353454


def test_sampler_2d_array(ctx):
    """RGBA8 2d array sampler. Read two pixels from two different layers"""
    prog = ctx.program(
        vertex_shader="""
        #version 330
        uniform sampler2DArray tex;
        out float color;
        void main() {
            color = texelFetch(tex, ivec3(gl_VertexID % 2, 0, gl_VertexID / 2), 0).r;
        }
        """,
        varyings=['color'],
    )
    tex = ctx.texture_array((2, 1, 2), 1, data=array('B', [127, 0, 255, 64]), dtype='f1')
    buff = ctx.buffer(reserve=4 * 4)
    vao = ctx.vertex_array(prog, [])
    tex.use(0)
    vao.transform(buff, vertices=4)
    data = struct.unpack('4f', buff.read())
    assert pytest.approx(data[0], abs=1.0e-3) == 0.498
    assert pytest.approx(data[1]) == 0.0
    assert pytest.approx(data[2]) == 1.0
    assert pytest.approx(data[3], abs=1.0e-3) == 0.25


def test_sampler_1d(ctx):
    prog = ctx.program(
        vertex_shader="""
        #version 330
        uniform sampler1D tex;
        out vec4 color;
        void main() {
            color = texelFetch(tex, 0, 0);
        }
        """,
        varyings=["color"],
    )
    assert "tex" in prog



import os
import moderngl


def test_local():
    path1 = os.path.normpath(os.path.dirname(os.path.dirname(moderngl.__file__)))
    path2 = os.path.normpath(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    assert path1.lower() == path2.lower(), 'Using installed moderngl'




def test_sampler_2D_get_setter(ctx):
    prog = ctx.program(
        vertex_shader='''
            #version 330
            uniform sampler2D Texture2D;
            void main() {
                gl_Position = texture(Texture2D, vec2(0.0, 0.0));
            }
        '''
    )

    sampler_uniform = prog['Texture2D']

    sampler_uniform.value = 3
    assert sampler_uniform.value == 3

    sampler_uniform.value = 1
    assert sampler_uniform.value == 1


def test_sampler_3D_get_setter(ctx):
    prog = ctx.program(
        vertex_shader='''
            #version 330
            uniform sampler3D Texture3D;
            void main() {
                gl_Position = texture(Texture3D, vec3(0.0, 0.0, 0.0));
            }
        '''
    )

    sampler_uniform = prog['Texture3D']

    sampler_uniform.value = 3
    assert sampler_uniform.value == 3

    sampler_uniform.value = 1
    assert sampler_uniform.value == 1



import moderngl


def test_create(ctx):
    size = (64, 32, 8)
    texture = ctx.texture_array(size, 4)

    # Basic attributes
    assert texture.size == size
    assert texture.width == 64
    assert texture.height == 32
    assert texture.layers == 8
    assert texture.dtype == 'f1'
    assert texture.components == 4
    assert texture.anisotropy == 0.0

    # Texture parameters
    assert texture.repeat_x is True
    assert texture.repeat_y is True
    assert texture.filter == (moderngl.LINEAR, moderngl.LINEAR)
    assert texture.swizzle == "RGBA"
    assert texture == texture

    texture.repeat_x = False
    texture.repeat_y = False
    assert texture.repeat_x is False
    assert texture.repeat_y is False

    assert texture.swizzle == "RGBA"
    texture.swizzle = "RGBA"
    assert texture.swizzle == "RGBA"

    texture.anisotropy = ctx.max_anisotropy
    assert texture.anisotropy == ctx.max_anisotropy

    assert texture.filter == (moderngl.LINEAR, moderngl.LINEAR)
    texture.filter = moderngl.NEAREST, moderngl.NEAREST
    assert texture.filter == (moderngl.NEAREST, moderngl.NEAREST)
    texture.build_mipmaps()
    assert texture.filter == (moderngl.LINEAR_MIPMAP_LINEAR, moderngl.LINEAR)


def test_texture_default_filter(ctx):
    """Ensure default filter is correct"""
    # Float types
    for dtype in ["f1", "f2", "f4"]:
        texture = ctx.texture_array((10, 10, 10), 4, dtype=dtype)
        assert texture.filter == (moderngl.LINEAR, moderngl.LINEAR)

    for dtype in ["u1", "u2", "u4", "i1", "i2", "i4"]:
        texture = ctx.texture_array((10, 10, 10), 4, dtype=dtype)
        assert texture.filter == (moderngl.NEAREST, moderngl.NEAREST)




def test_1(ctx):
    buf1 = ctx.buffer(data=b'\xAA\x55' * 10)
    buf2 = ctx.buffer(reserve=buf1.size)
    buf2.write(b'Hello World!')

    assert buf1.read(5, offset=1) == b'\x55\xaa\x55\xaa\x55'
    assert buf2.read(5, offset=6) == b'World'

    ctx.copy_buffer(buf2, buf1, read_offset=1, write_offset=6, size=5)
    assert buf2.read(12) == b'Hello \x55\xaa\x55\xaa\x55!'





def test_glo(ctx):
    buf = ctx.buffer(reserve=1024)
    assert buf.glo > 0


def test_buffer_eq(ctx):
    buf1 = ctx.buffer(reserve=1024)
    buf2 = ctx.buffer(reserve=1024)
    assert buf1 != buf2
    assert buf1 == buf1
    assert buf2 == buf2


def test_bind(ctx):
    buf = ctx.buffer(reserve=1024)
    res = buf.bind("in_pos", "in_vel", layout="2f 2f")
    assert res == (buf, "2f 2f", "in_pos", "in_vel")


def test_assign(ctx):
    buf = ctx.buffer(reserve=1024)
    assert buf.assign(1) == (buf, 1)


def test_buffer_clear_1(ctx):
    buf = ctx.buffer(data=b'\xAA\x55' * 10)
    buf.clear(chunk=b'AB')
    assert buf.read() == b'AB' * 10


def test_buffer_clear_2(ctx):
    buf = ctx.buffer(data=b'\xAA\x55' * 10)
    buf.clear(offset=1, size=18, chunk=b'AB')
    assert buf.read() == b'\xAAABABABABABABABABAB\x55'


def test_buffer_create(ctx):
    buf = ctx.buffer(data=b'\xAA\x55' * 10)
    assert buf.read() == b'\xAA\x55' * 10


def test_buffer_read_write(ctx):
    buf = ctx.buffer(reserve=10)
    buf.write(b'abcd')
    assert buf.read(4) == b'abcd'
    buf.write(b'abcd', offset=3)
    assert buf.read(4, offset=1) == b'bcab'
    buf.write(b'abcd', offset=6)
    assert buf.read() == b'abcabcabcd'
    assert buf.read(offset=3) == b'abcabcd'


def test_buffer_read_into_1(ctx):
    data = b'Hello World!'
    buf = ctx.buffer(data)
    res = bytearray(buf.size)
    buf.read_into(res)
    assert bytes(res) == data


def test_buffer_read_into_2(ctx):
    data = b'Hello World!'
    buf = ctx.buffer(data)
    res = bytearray(10)
    buf.read_into(res, offset=6, size=5, write_offset=0)
    buf.read_into(res, offset=0, size=5, write_offset=5)
    assert bytes(res) == b'WorldHello'


def test_buffer_orphan(ctx):
    buf = ctx.buffer(reserve=1024)
    buf.orphan()


def test_buffer_orphan_resize(ctx):
    buf = ctx.buffer(reserve=10)
    assert buf.size == 10
    assert len(buf.read()) == 10

    buf.orphan(20)
    assert buf.size == 20
    assert len(buf.read()) == 20

    buf.orphan(100)
    assert buf.size == 100
    assert len(buf.read()) == 100

    buf.orphan(-100)
    assert buf.size == 100
    assert len(buf.read()) == 100



import struct
import pytest
import moderngl


def test_1(ctx):
    prog = ctx.program(
        vertex_shader='''
            #version 330

            in vec2 in_v1;
            in vec2 in_v2;

            out vec2 out_v;

            void main() {
                out_v = in_v1 + in_v2;
            }
        ''',
        varyings=['out_v']
    )

    buf_v1 = ctx.buffer(struct.pack('8f', 1, 2, 3, 4, 5, 6, 7, 8))
    buf_v2 = ctx.buffer(struct.pack('2f', 10, 100))
    res = ctx.buffer(reserve=buf_v1.size)

    vao = ctx.vertex_array(prog, [
        (buf_v1, '2f', 'in_v1'),
        (buf_v2, '2f/i', 'in_v2'),
    ])

    vao.transform(res, moderngl.POINTS)
    tup = struct.unpack('8f', res.read())
    assert pytest.approx(tup[0]) == 11.0
    assert pytest.approx(tup[1]) == 102.0
    assert pytest.approx(tup[2]) == 13.0
    assert pytest.approx(tup[3]) == 104.0
    assert pytest.approx(tup[4]) == 15.0
    assert pytest.approx(tup[5]) == 106.0
    assert pytest.approx(tup[6]) == 17.0
    assert pytest.approx(tup[7]) == 108.0



"""
Test that all objects can be released.
"""


def test_texture(ctx):
    tex = ctx.texture((4, 4), 4)
    tex.release()


def test_texture_array(ctx):
    tex = ctx.texture_array((4, 4, 4), 4)
    tex.release()


def test_texture_cube(ctx):
    tex = ctx.texture_cube((4, 4), 4)
    tex.release()


def test_texture_3d(ctx):
    tex = ctx.texture3d((4, 4, 4), 4)
    tex.release()


def test_renderbuffer(ctx):
    rbo = ctx.renderbuffer((4, 4))
    rbo.release()


def test_buffer(ctx):
    buf = ctx.buffer(reserve=4)
    buf.release()


def test_framebuffer(ctx):
    fbo = ctx.framebuffer(ctx.renderbuffer((4, 4)))
    fbo.release()


def test_vertex_array(ctx):
    vao = ctx.vertex_array(ctx.program(
        vertex_shader="""
            #version 330
            void main() {
                gl_Position = vec4(0.0, 0.0, 0.0, 1.0);
            }
        """,
    ), [])
    vao.release()


def test_sampler(ctx):
    smp = ctx.sampler()
    smp.release()


def test_scope(ctx):
    fbo = ctx.simple_framebuffer((4, 4))
    scope = ctx.scope(framebuffer=fbo)
    fbo.release()
    scope.release()



"""
Test transform feedback.

Note that some of the tests are commented out because they currently don't
work with mesa/lvvmpipe. They are left here for future reference.
"""
from array import array
import struct
import numpy as np
import moderngl


def test_transform(ctx):

    program = ctx.program(
        vertex_shader='''
            #version 330

            in vec4 vert;
            out float vert_length;

            void main() {
                vert_length = length(vert);
            }
        ''',
        varyings=['vert_length']
    )

    assert program.geometry_input is None
    assert program.geometry_output is None
    assert program.is_transform is True

    assert 'vert' in program
    assert 'vert_length' in program


def test_vertex_points(ctx):
    """Transform with points"""
    print(ctx.error)
    data = (0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0)
    buffer = ctx.buffer(array('f', data))
    program = ctx.program(
        vertex_shader="""
        #version 330

        in vec2 in_pos;
        out vec2 out_pos;

        void main() {
            out_pos = in_pos;
        }
        """,
        varyings=["out_pos"],
    )

    vao = ctx.vertex_array(program, [(buffer, "2f", "in_pos"), ])
    assert vao.mode == moderngl.POINTS
    buffer2 = ctx.buffer(reserve=buffer.size)
    vao.transform(buffer2, mode=ctx.POINTS)
    assert data == struct.unpack(f"{len(data)}f", buffer.read())


def test_vertex_lines(ctx):
    data = 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0
    buffer = ctx.buffer(array('f', data))
    program = ctx.program(
        vertex_shader="""
        #version 330

        in vec2 in_pos;
        out vec2 out_pos;

        void main() {
            out_pos = in_pos * 2.0;
        }
        """,
        varyings=["out_pos"],
    )

    vao = ctx.vertex_array(program, [(buffer, "2f", "in_pos"), ])
    buffer2 = ctx.buffer(reserve=buffer.size)
    vao.transform(buffer2, mode=ctx.LINES)
    assert tuple(v * 2 for v in data) == struct.unpack(f"{len(data)}f", buffer2.read())


# NOTE: This fails in tests and may be non-standard or require higher gl version
# def test_vertex_lines_adjacency(ctx):
#     if platform.system().lower() in ["darwin"]:
#         self.skipTest('Transforms with adjacency primitives now working on OSX')

#     data = (
#         1.0, 1.0,
#         2.0, 2.0,
#         3.0, 3.0,
#         4.0, 4.0,
#     )
#     buffer = ctx.buffer(array('f', data))
#     program = ctx.program(
#         vertex_shader="""
#         #version 330

#         in vec2 in_pos;
#         out vec2 out_pos;

#         void main() {
#             out_pos = in_pos;
#         }
#         """,
#         varyings=["out_pos"],
#     )
#     vao = ctx.vertex_array(program, [(buffer, "2f", "in_pos"),])
#     buffer2 = ctx.buffer(reserve=4 * 4)
#     vao.transform(buffer2, mode=ctx.LINES_ADJACENCY)

#     # Start end end line is removed
#     self.assertEqual(tuple(data[2:6]), struct.unpack(f"4f", buffer2.read()))


def test_vertex_line_strip(ctx):
    data = (
        1.0, 1.0,
        2.0, 2.0,
        3.0, 3.0,
        4.0, 4.0,
    )
    buffer = ctx.buffer(array('f', data))
    program = ctx.program(
        vertex_shader="""
        #version 330

        in vec2 in_pos;
        out vec2 out_pos;

        void main() {
            out_pos = in_pos;
        }
        """,
        varyings=["out_pos"],
    )
    vao = ctx.vertex_array(program, [(buffer, "2f", "in_pos"), ])
    buffer2 = ctx.buffer(reserve=4 * 12)
    vao.transform(buffer2, mode=ctx.LINE_STRIP)

    expected = (1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 4.0, 4.0)
    assert expected == struct.unpack("12f", buffer2.read())


# NOTE: Fails in tests and might be non-standard or require higher gl version
# def test_vertex_line_strip_adjacency(ctx):
#     if platform.system().lower() in ["darwin"]:
#         self.skipTest('Transforms with adjacency primitives now working on OSX')

#     data = (
#         1.0, 1.0,
#         2.0, 2.0,
#         3.0, 3.0,
#         4.0, 4.0,
#         5.0, 5.0,
#     )
#     buffer = ctx.buffer(array('f', data))
#     program = ctx.program(
#         vertex_shader="""
#         #version 330

#         in vec2 in_pos;
#         out vec2 out_pos;

#         void main() {
#             out_pos = in_pos;
#         }
#         """,
#         varyings=["out_pos"],
#     )
#     vao = ctx.vertex_array(program, [(buffer, "2f", "in_pos"),])
#     buffer2 = ctx.buffer(reserve=4 * 8)
#     vao.transform(buffer2, mode=ctx.LINE_STRIP_ADJACENCY)
#     self.assertEqual(
#         (2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 4.0, 4.0),
#         struct.unpack("8f", buffer2.read())
#     )


def test_vertex_triangles(ctx):
    data = (
        1.0, 1.0,
        2.0, 2.0,
        3.0, 3.0,
        4.0, 4.0,
        5.0, 5.0,
        6.0, 6.0,
    )
    buffer = ctx.buffer(array('f', data))
    program = ctx.program(
        vertex_shader="""
        #version 330

        in vec2 in_pos;
        out vec2 out_pos;

        void main() {
            out_pos = in_pos * 2;
        }
        """,
        varyings=["out_pos"],
    )
    vao = ctx.vertex_array(program, [(buffer, "2f", "in_pos"), ])
    buffer2 = ctx.buffer(reserve=4 * 12)
    vao.transform(buffer2, mode=ctx.TRIANGLES)
    assert tuple(v * 2 for v in data) == struct.unpack("12f", buffer2.read())


# def test_vertex_triangles_adjacency(ctx):
#     if platform.system().lower() in ["darwin"]:
#         pytest.skip('Transforms with adjacency primitives now working on OSX')
#     data = (
#         1.0, 1.0, # outer points
#         2.0, 2.0,
#         3.0, 3.0,
#         4.0, 4.0,
#         5.0, 5.0,
#         6.0, 6.0,
#         7.0, 7.0, # inner triangle
#         8.0, 8.0,
#         9.0, 9.0,
#     )
#     buffer = ctx.buffer(array('f', data))
#     program = ctx.program(
#         vertex_shader="""
#         #version 330

#         in vec2 in_pos;
#         out vec2 out_pos;

#         void main() {
#             out_pos = in_pos;
#         }
#         """,
#         varyings=["out_pos"],
#     )
#     vao = ctx.vertex_array(program, [(buffer, "2f", "in_pos"),])
#     buffer2 = ctx.buffer(reserve=4 * 6)
#     vao.transform(buffer2, mode=ctx.TRIANGLES_ADJACENCY)
#     assert (1.0, 1.0, 3.0, 3.0, 5.0, 5.0) == struct.unpack("6f", buffer2.read())

# def test_vertex_triangle_strip(ctx):
#     data = (
#         1.0, 1.0,
#         2.0, 2.0,
#         3.0, 3.0,
#         4.0, 4.0,
#     )
#     buffer = ctx.buffer(array('f', data))
#     program = ctx.program(
#         vertex_shader="""
#         #version 330

#         in vec2 in_pos;
#         out vec2 out_pos;

#         void main() {
#             out_pos = in_pos;
#         }
#         """,
#         varyings=["out_pos"],
#     )
#     vao = ctx.vertex_array(program, [(buffer, "2f", "in_pos"),])
#     buffer2 = ctx.buffer(reserve=4 * 12)
#     vao.transform(buffer2, mode=ctx.TRIANGLE_STRIP)
#     expected = (1.0, 1.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 2.0, 2.0, 4.0, 4.0),
#     assert expected == struct.unpack("12f", buffer2.read())

# def test_vertex_triangle_fan(ctx):
#     vertices = np.array(
#         [
#             0.0, 0.0,  # Initial triangle
#             1.0, 0.0,
#             1.0, 1.0,
#             0.0, 1.0,  # second triangle
#         ],
#         dtype='f4',
#     )

#     prog = ctx.program(vertex_shader=
#         """
#         #version 330
#         in vec2 in_vert;
#         out vec2 out_vert;
#         void main(void) {
#             out_vert = in_vert;
#         }
#         """,
#         varyings=['out_vert'],
#     )

#     vbo = ctx.buffer(vertices)
#     vao = ctx.vertex_array(prog, [(vbo, '2f', 'in_vert')])
#     N = 6  # 6 vertices in end buffer
#     buffer = ctx.buffer(reserve=N * 4 * 2)
#     vao.transform(buffer, vertices=4, mode=ctx.TRIANGLE_FAN)

#     data = struct.unpack('{}f'.format(N * 2), buffer.read())
#     expected = (
#         0.0, 0.0,  # Initial triangle
#         1.0, 0.0,
#         1.0, 1.0,
#         0.0, 0.0,  # second triangle
#         1.0, 1.0,
#         0.0, 1.0,
#     )
#     assert data == expected

# # ------ Geometry shader -------

def test_geometry_points(ctx):
    data = (1.0, 2.0, 3.0, 4.0, 5.0, 6.0)
    vertices = np.array(data, dtype='f4')
    prog = ctx.program(
        vertex_shader="""
        #version 330

        in float in_value;

        void main() {
            gl_Position = vec4(in_value);
        }
        """,
        geometry_shader="""
        #version 330

        out float out_value;

        layout(points) in;
        layout(points, max_vertices=1) out;

        void main() {
            out_value = gl_in[0].gl_Position.x * 2;
            EmitVertex();
        }
        """,
        varyings=['out_value'],
    )
    vbo = ctx.buffer(vertices)
    vao = ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
    buffer = ctx.buffer(reserve=vbo.size)
    vao.transform(buffer, mode=moderngl.POINTS)
    assert tuple(v * 2 for v in data) == struct.unpack('6f', buffer.read())


def test_geometry_lines(ctx):
    data = (
        1.0, 2.0,
        3.0, 4.0,
        5.0, 6.0,
    )
    vertices = np.array(data, dtype='f4')
    prog = ctx.program(
        vertex_shader="""
        #version 330

        in float in_value;

        void main() {
            gl_Position = vec4(in_value);
        }
        """,
        geometry_shader="""
        #version 330

        out float out_value;

        layout(lines) in;
        layout(line_strip, max_vertices=2) out;

        void main() {
            out_value = gl_in[0].gl_Position.x;
            EmitVertex();
            out_value = gl_in[1].gl_Position.x;
            EmitVertex();
            EndPrimitive();
        }
        """,
        varyings=['out_value'],
    )
    vbo = ctx.buffer(vertices)
    vao = ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
    buffer = ctx.buffer(reserve=vbo.size)
    vao.transform(buffer, mode=moderngl.LINES)
    assert data == struct.unpack('6f', buffer.read())


# def test_geometry_line_loop(ctx):
#     data = (
#         1.0, 2.0, 3.0, 4.0, 5.0, 6.0,
#     )
#     vertices = np.array(data, dtype='f4')
#     prog = ctx.program(
#         vertex_shader=
#         """
#         #version 330

#         in float in_value;

#         void main() {
#             gl_Position = vec4(in_value);
#         }
#         """,
#         geometry_shader="""
#         #version 330

#         out float out_value;

#         layout(lines) in;
#         layout(line_strip, max_vertices=2) out;

#         void main() {
#             out_value = gl_in[0].gl_Position.x;
#             EmitVertex();
#             out_value = gl_in[1].gl_Position.x;
#             EmitVertex();
#             EndPrimitive();
#         }
#         """,
#         varyings=['out_value'],
#     )
#     vbo = ctx.buffer(vertices)
#     vao = ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
#     buffer = ctx.buffer(reserve=4 * 12)
#     vao.transform(buffer, mode=moderngl.LINE_LOOP)
#     expected = (1.0, 2.0, 2.0, 3.0, 3.0, 4.0, 4.0, 5.0, 5.0, 6.0, 6.0, 1.0),
#     assert struct.unpack('12f', buffer.read()) == expected

# def test_geometry_line_strip(ctx):
#     data = (
#         1.0, 2.0, 3.0, 4.0, 5.0
#     )
#     vertices = np.array(data, dtype='f4')
#     prog = ctx.program(
#         vertex_shader=
#         """
#         #version 330

#         in float in_value;

#         void main() {
#             gl_Position = vec4(in_value);
#         }
#         """,
#         geometry_shader="""
#         #version 330

#         out float out_value;

#         layout(lines) in;
#         layout(line_strip, max_vertices=2) out;

#         void main() {
#             out_value = gl_in[0].gl_Position.x;
#             EmitVertex();
#             out_value = gl_in[1].gl_Position.x;
#             EmitVertex();
#             EndPrimitive();
#         }
#         """,
#         varyings=['out_value'],
#     )
#     vbo = ctx.buffer(vertices)
#     vao = ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
#     buffer = ctx.buffer(reserve=4 * 8)
#     vao.transform(buffer, mode=moderngl.LINE_LOOP)
#     expected = (1.0, 2.0, 2.0, 3.0, 3.0, 4.0, 4.0, 5.0),
#     assert struct.unpack('8f', buffer.read()) == expected

# def test_geometry_lines_adjacency(ctx):
#     data = (
#         1.0, 2.0, 3.0, 4.0,
#     )
#     vertices = np.array(data, dtype='f4')
#     prog = ctx.program(
#         vertex_shader=
#         """
#         #version 330

#         in float in_value;

#         void main() {
#             gl_Position = vec4(in_value);
#         }
#         """,
#         geometry_shader="""
#         #version 330

#         out float out_value;

#         layout(lines_adjacency) in;
#         layout(line_strip, max_vertices=2) out;

#         void main() {
#             out_value = gl_in[1].gl_Position.x;
#             EmitVertex();
#             out_value = gl_in[2].gl_Position.x;
#             EmitVertex();
#             EndPrimitive();
#         }
#         """,
#         varyings=['out_value'],
#     )
#     vbo = ctx.buffer(vertices)
#     vao = ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
#     buffer = ctx.buffer(reserve=4 * 2)
#     vao.transform(buffer, mode=moderngl.LINES_ADJACENCY)
#     assert struct.unpack('2f', buffer.read()) == (2.0, 3.0)

# def test_geometry_line_strip_adjacency(ctx):
#     data = (
#         1.0, 2.0, 3.0, 4.0, 5.0,
#     )
#     vertices = np.array(data, dtype='f4')
#     prog = ctx.program(
#         vertex_shader=
#         """
#         #version 330

#         in float in_value;

#         void main() {
#             gl_Position = vec4(in_value);
#         }
#         """,
#         geometry_shader="""
#         #version 330

#         out float out_value;

#         layout(lines_adjacency) in;
#         layout(line_strip, max_vertices=2) out;

#         void main() {
#             out_value = gl_in[1].gl_Position.x;
#             EmitVertex();
#             out_value = gl_in[2].gl_Position.x;
#             EmitVertex();
#             EndPrimitive();
#         }
#         """,
#         varyings=['out_value'],
#     )
#     vbo = ctx.buffer(vertices)
#     vao = ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
#     buffer = ctx.buffer(reserve=4 * 4)
#     vao.transform(buffer, mode=moderngl.LINE_STRIP_ADJACENCY)       
#     assert struct.unpack('4f', buffer.read()) == (2.0, 3.0, 3.0, 4.0)


def test_geometry_triangles(ctx):
    data = (
        1.0, 2.0, 3.0, 4.0, 5.0, 6.0,
    )
    vertices = np.array(data, dtype='f4')
    prog = ctx.program(
        vertex_shader="""
        #version 330

        in float in_value;

        void main() {
            gl_Position = vec4(in_value);
        }
        """,
        geometry_shader="""
        #version 330

        out float out_value;

        layout(triangles) in;
        layout(triangle_strip, max_vertices=3) out;

        void main() {
            out_value = gl_in[0].gl_Position.x;
            EmitVertex();
            out_value = gl_in[1].gl_Position.x;
            EmitVertex();
            out_value = gl_in[2].gl_Position.x;
            EmitVertex();
            EndPrimitive();
        }
        """,
        varyings=['out_value'],
    )
    vbo = ctx.buffer(vertices)
    vao = ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
    buffer = ctx.buffer(reserve=vbo.size)
    vao.transform(buffer, mode=moderngl.TRIANGLES)
    assert struct.unpack('6f', buffer.read()) == data


# def test_geometry_triangle_strip(ctx):
#     data = (
#         1.0, 2.0, 3.0, 4.0,
#     )
#     vertices = np.array(data, dtype='f4')
#     prog = ctx.program(
#         vertex_shader=
#         """
#         #version 330

#         in float in_value;

#         void main() {
#             gl_Position = vec4(in_value);
#         }
#         """,
#         geometry_shader="""
#         #version 330

#         out float out_value;

#         layout(triangles) in;
#         layout(triangle_strip, max_vertices=3) out;

#         void main() {
#             out_value = gl_in[0].gl_Position.x;
#             EmitVertex();
#             out_value = gl_in[1].gl_Position.x;
#             EmitVertex();
#             out_value = gl_in[2].gl_Position.x;
#             EmitVertex();
#             EndPrimitive();
#         }
#         """,
#         varyings=['out_value'],
#     )
#     vbo = ctx.buffer(vertices)
#     vao = ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
#     buffer = ctx.buffer(reserve=4 * 6)
#     vao.transform(buffer, mode=moderngl.TRIANGLE_STRIP)
#     expected = (1.0, 2.0, 3.0, 3.0, 2.0, 4.0)
#     assert struct.unpack('6f', buffer.read()) == expected

# def test_geometry_triangle_fan(ctx):
#     data = (
#         0.0, 1.0, 2.0, 3.0,
#     )
#     vertices = np.array(data, dtype='f4')
#     prog = ctx.program(
#         vertex_shader=
#         """
#         #version 330

#         in float in_value;

#         void main() {
#             gl_Position = vec4(in_value);
#         }
#         """,
#         geometry_shader="""
#         #version 330

#         out float out_value;

#         layout(triangles) in;
#         layout(triangle_strip, max_vertices=3) out;

#         void main() {
#             out_value = gl_in[0].gl_Position.x;
#             EmitVertex();
#             out_value = gl_in[1].gl_Position.x;
#             EmitVertex();
#             out_value = gl_in[2].gl_Position.x;
#             EmitVertex();
#             EndPrimitive();
#         }
#         """,
#         varyings=['out_value'],
#     )
#     vbo = ctx.buffer(vertices)
#     vao = ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
#     buffer = ctx.buffer(reserve=4 * 6)
#     vao.transform(buffer, mode=moderngl.TRIANGLE_FAN)
#     expected = (0.0, 1.0, 2.0, 0.0, 2.0, 3.0)
#     assert struct.unpack('6f', buffer.read()) == expected


# def test_geometry_triangles_adjacency(ctx):
#     data = (
#         1.0, 2.0, 3.0, 4.0, 5.0, 6.0,
#     )
#     vertices = np.array(data, dtype='f4')
#     prog = ctx.program(
#         vertex_shader=
#         """
#         #version 330

#         in float in_value;

#         void main() {
#             gl_Position = vec4(in_value);
#         }
#         """,
#         geometry_shader="""
#         #version 330

#         out float out_value;

#         layout(triangles_adjacency) in;
#         layout(triangle_strip, max_vertices=3) out;

#         void main() {
#             out_value = gl_in[0].gl_Position.x;
#             EmitVertex();
#             out_value = gl_in[2].gl_Position.x;
#             EmitVertex();
#             out_value = gl_in[4].gl_Position.x;
#             EmitVertex();
#             EndPrimitive();
#         }
#         """,
#         varyings=['out_value'],
#     )
#     vbo = ctx.buffer(vertices)
#     vao = ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
#     buffer = ctx.buffer(reserve=4 * 3)
#     vao.transform(buffer, mode=moderngl.TRIANGLES_ADJACENCY)
#     assert struct.unpack('3f', buffer.read()) == (1.0, 3.0, 5.0)


# def test_geometry_triangle_strip_adjacency(ctx):
#     data = (
#         1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0,
#     )
#     vertices = np.array(data, dtype='f4')
#     prog = ctx.program(
#         vertex_shader=
#         """
#         #version 330

#         in float in_value;

#         void main() {
#             gl_Position = vec4(in_value);
#         }
#         """,
#         geometry_shader="""
#         #version 330

#         out float out_value;

#         layout(triangles_adjacency) in;
#         layout(triangle_strip, max_vertices=3) out;

#         void main() {
#             out_value = gl_in[0].gl_Position.x;
#             EmitVertex();
#             out_value = gl_in[2].gl_Position.x;
#             EmitVertex();
#             out_value = gl_in[4].gl_Position.x;
#             EmitVertex();
#             EndPrimitive();
#         }
#         """,
#         varyings=['out_value'],
#     )
#     vbo = ctx.buffer(vertices)
#     vao = ctx.vertex_array(prog, [(vbo, 'f', 'in_value')])
#     buffer = ctx.buffer(reserve=4 * 6)
#     vao.transform(buffer, mode=moderngl.TRIANGLE_STRIP_ADJACENCY)

#     expected = (1.0, 3.0, 5.0, 5.0, 3.0, 7.0)
#     assert struct.unpack('6f', buffer.read()) == expected



import moderngl
import pytest


def test_framebuffer_color_attachment(ctx):
    rbo = ctx.renderbuffer((16, 16))
    ctx.framebuffer(rbo)


def test_framebuffer_get_color_attachments(ctx):
    rbo1 = ctx.renderbuffer((16, 16))
    rbo2 = ctx.renderbuffer((16, 16))
    rbo3 = ctx.renderbuffer((16, 16))

    fbo1 = ctx.framebuffer(rbo1)
    fbo2 = ctx.framebuffer([rbo2, rbo1])
    fbo3 = ctx.framebuffer([rbo1, rbo2, rbo3])

    assert len(fbo1.color_attachments) == 1
    assert len(fbo2.color_attachments) == 2
    assert len(fbo3.color_attachments) == 3

    assert isinstance(fbo1.color_attachments[0], moderngl.Renderbuffer)
    assert isinstance(fbo2.color_attachments[0], moderngl.Renderbuffer)
    assert isinstance(fbo2.color_attachments[1], moderngl.Renderbuffer)
    assert isinstance(fbo3.color_attachments[0], moderngl.Renderbuffer)
    assert isinstance(fbo3.color_attachments[1], moderngl.Renderbuffer)
    assert isinstance(fbo3.color_attachments[2], moderngl.Renderbuffer)

    assert rbo1 in fbo1.color_attachments
    assert rbo2 in fbo2.color_attachments
    assert rbo2 in fbo3.color_attachments

    assert rbo2 not in fbo1.color_attachments
    assert rbo3 not in fbo1.color_attachments
    assert rbo3 not in fbo2.color_attachments


def test_framebuffer_get_depth_attachment(ctx):
    rbo1 = ctx.renderbuffer((16, 16))
    rbo2 = ctx.depth_renderbuffer((16, 16))

    fbo1 = ctx.framebuffer(rbo1)
    fbo2 = ctx.framebuffer(rbo1, rbo2)

    assert fbo1.depth_attachment is None
    assert isinstance(fbo2.depth_attachment, moderngl.Renderbuffer)

    assert fbo1.depth_attachment != rbo2
    assert fbo2.depth_attachment == rbo2


def test_framebuffer_color_mask(ctx):
    fbo = ctx.framebuffer(ctx.renderbuffer((16, 16)))
    assert fbo.color_mask == (True, True, True, True)


def test_framebuffer_single_channel_color_mask(ctx):
    fbo = ctx.framebuffer(ctx.renderbuffer((16, 16), components=1))
    assert fbo.color_mask == (True, False, False, False)


def test_framebuffer_mixed_channels_color_mask(ctx):
    fbo = ctx.framebuffer([
        ctx.renderbuffer((16, 16), components=1),
        ctx.renderbuffer((16, 16), components=2),
        ctx.renderbuffer((16, 16), components=3),
        ctx.renderbuffer((16, 16), components=4),
    ])
    expected = (
        (True, False, False, False),
        (True, True, False, False),
        (True, True, True, False),
        (True, True, True, True),
    )
    assert fbo.color_mask == expected


def test_framebuffer_depth_mask_1(ctx):
    fbo = ctx.framebuffer(
        ctx.renderbuffer((16, 16)),
        ctx.depth_renderbuffer((16, 16)),
    )
    assert fbo.depth_mask is True


def test_framebuffer_depth_mask_2(ctx):
    fbo = ctx.framebuffer(
        ctx.renderbuffer((16, 16))
    )
    assert fbo.depth_mask is False


def test_framebuffer_color_attachments(ctx):
    rbo1 = ctx.renderbuffer((16, 16))
    rbo2 = ctx.renderbuffer((16, 16))
    rbo3 = ctx.renderbuffer((16, 16))
    ctx.framebuffer([rbo1, rbo2, rbo3])


def test_framebuffer_multiple_color_masks(ctx):
    fbo = ctx.framebuffer([
        ctx.renderbuffer((16, 16)),
        ctx.renderbuffer((16, 16)),
    ])
    expected = ((True, True, True, True), (True, True, True, True))
    assert fbo.color_mask == expected


def test_framebuffer_size_mismatch(ctx):
    with pytest.raises(moderngl.Error, match='size'):
        rbo1 = ctx.renderbuffer((16, 16))
        rbo2 = ctx.depth_renderbuffer((32, 32))
        ctx.framebuffer(rbo1, rbo2)


def test_framebuffer_color_attachments_size_mismatch(ctx):
    with pytest.raises(moderngl.Error, match='size'):
        rbo1 = ctx.renderbuffer((16, 16))
        rbo2 = ctx.renderbuffer((32, 32))
        ctx.framebuffer([rbo1, rbo2])


def test_depth_framebuffer(ctx):
    rbo1 = ctx.renderbuffer((16, 16))
    rbo2 = ctx.depth_renderbuffer((16, 16))
    ctx.framebuffer(rbo1, rbo2)


def test_framebuffer_multisample(ctx):
    if ctx.max_samples < 2:
        pytest.skip('multisampling is not supported')

    rbo1 = ctx.renderbuffer((16, 16), samples=2)
    ctx.framebuffer(rbo1)


def test_depth_framebuffer_multisample(ctx):
    if ctx.max_samples < 2:
        pytest.skip('multisampling is not supported')

    rbo1 = ctx.renderbuffer((16, 16), samples=2)
    rbo2 = ctx.depth_renderbuffer((16, 16), samples=2)
    ctx.framebuffer(rbo1, rbo2)


def test_framebuffer_multisample_sample_mismatch(ctx):
    if ctx.max_samples < 2:
        pytest.skip('multisampling is not supported')

    with pytest.raises(moderngl.Error, match='sample'):
        rbo1 = ctx.renderbuffer((16, 16))
        rbo2 = ctx.depth_renderbuffer((16, 16), samples=2)
        ctx.framebuffer(rbo1, rbo2)


def test_empty_framebuffer(ctx):
    with pytest.raises(moderngl.Error, match='empty'):
        ctx.framebuffer([])


def test_framebuffer_having_depth_in_colors(ctx):
    with pytest.raises(moderngl.Error, match=r'(color|depth)'):
        ctx.framebuffer(ctx.depth_renderbuffer((16, 16)))


def test_framebuffer_having_color_in_depth(ctx):
    with pytest.raises(moderngl.Error, match=r'(color|depth)'):
        ctx.framebuffer(
            ctx.renderbuffer((16, 16)),
            ctx.renderbuffer((16, 16)),
        )



import struct
import pytest


def test_1(ctx):
    if ctx.version_code < 430:
        raise pytest.skip('Compute shaders not supported')

    compute_shader = ctx.compute_shader('''
        #version 430

        layout (local_size_x = 1, local_size_y = 1) in;

        layout (std430, binding = 1) buffer Input {
            float v1[4];
        };

        layout (std430, binding = 2) buffer Output {
            float v2[4];
        };

        uniform float mul;
        uniform vec4 add;

        void main() {
            v2[0] = v1[3] * mul + add.x;
            v2[1] = v1[2] * mul + add.y;
            v2[2] = v1[1] * mul + add.z;
            v2[3] = v1[0] * mul + add.w;
        }
    ''')

    buf1 = ctx.buffer(struct.pack('4f', 1.0, 2.0, 3.0, 4.0))
    buf2 = ctx.buffer(struct.pack('4f', 0.0, 0.0, 0.0, 0.0))

    buf1.bind_to_storage_buffer(1)
    buf2.bind_to_storage_buffer(2)

    compute_shader['add'].value = (10.5, 20.0, 30.5, 40.0)
    compute_shader['mul'].value = 100.0
    compute_shader.run()

    a, b, c, d = struct.unpack('4f', buf2.read())

    assert pytest.approx(a) == 410.5
    assert pytest.approx(b) == 320.0
    assert pytest.approx(c) == 230.5
    assert pytest.approx(d) == 140.0



import os
import platform
import re
import sys

from setuptools import Extension, setup

PLATFORMS = {"windows", "linux", "darwin", "cygwin", "android"}

target = platform.system().lower()

if "pydroid3" in sys.executable.lower():
    target = "android"

for known in PLATFORMS:
    if target.startswith(known):
        target = known

if target not in PLATFORMS:
    target = "linux"

libraries = {
    "windows": [],
    "linux": [],
    "cygwin": [],
    "darwin": [],
    "android": [],
}

extra_compile_args = {
    "windows": ["-fpermissive"],
    "linux": ["-fpermissive"],
    "cygwin": ["-fpermissive"],
    "darwin": ["-Wno-deprecated-declarations"],
    "android": ["-fpermissive"],
}

extra_linker_args = {
    "windows": [],
    "linux": [],
    "cygwin": [],
    "darwin": [],
    "android": [],
}

if os.getenv("MODERNGL_COVERAGE"):
    extra_compile_args[target] += ["-O0", "--coverage"]
    extra_linker_args[target] += ["-O0", "--coverage"]

mgl = Extension(
    name="moderngl.mgl",
    libraries=libraries[target],
    extra_compile_args=extra_compile_args[target],
    extra_link_args=extra_linker_args[target],
    define_macros=[('PY_SSIZE_T_CLEAN', None)],
    sources=["src/moderngl.cpp"],
)

short_description = "ModernGL: High performance rendering for Python 3"

with open("README.md") as f:
    long_description = re.sub(r"</?div[^>]*>|\r", "", f.read(), flags=re.M)

keywords = [
    "ModernGL",
    "OpenGL",
    "PyOpenGL",
    "rendering",
    "graphics",
    "shader",
    "GLSL",
    "GPU",
    "visualization",
    "2D",
    "3D",
]

classifiers = [
    "Development Status :: 5 - Production/Stable",
    "License :: OSI Approved",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Topic :: Games/Entertainment",
    "Topic :: Multimedia :: Graphics",
    "Topic :: Multimedia :: Graphics :: 3D Rendering",
    "Topic :: Scientific/Engineering :: Visualization",
    "Programming Language :: Python :: 3 :: Only",
]

project_urls = {
    "Documentation": "https://moderngl.readthedocs.io/",
    "Source": "https://github.com/moderngl/moderngl/",
    "Bug Tracker": "https://github.com/moderngl/moderngl/issues/",
}

setup(
    name="moderngl",
    version="5.9.0",
    description=short_description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/moderngl/moderngl",
    author="Szabolcs Dombi",
    author_email="szabolcs@szabolcsdombi.com",
    license="MIT",
    project_urls=project_urls,
    classifiers=classifiers,
    keywords=keywords,
    include_package_data=True,
    package_data={"moderngl-stubs": ["__init__.pyi"]},
    packages=["moderngl", "moderngl-stubs"],
    py_modules=["_moderngl"],
    ext_modules=[mgl],
    platforms=["any"],
    install_requires=["glcontext>=2.5.0,<3"],
    python_requires=">=3.7",
)



import moderngl
import pytest


def test_compute_shader(ctx):
    if ctx.version_code < 430:
        pytest.skip('OpenGL 4.3 is not supported')

    src = '''
        #version 430
        layout (local_size_x = 1, local_size_y = 1, local_size_z = 1) in;

        struct Something
        {
            vec4 pos;
        };

        layout(std430, binding=0) buffer something_in
        {
            Something entries[];
        } In;

        // output into the buffer bound to 1
        layout(std430, binding=1) buffer something_out
        {
            Something entries[];
        } Out;

        uniform float multiplier;

        void main() {
            int x = int(gl_GlobalInvocationID);
            Something entry = In.entries[x];

            Out.entries[x] = Something(entry.pos * multiplier);
        }
    '''

    buffer_in = ctx.buffer(reserve=16)
    buffer_out = ctx.buffer(reserve=16)

    buffer_in.bind_to_storage_buffer(binding=0)
    buffer_out.bind_to_storage_buffer(binding=1)

    compute_shader = ctx.compute_shader(src)

    assert isinstance(compute_shader.glo, int)
    assert compute_shader.mglo is not None
    compute_shader.extra = {'name': 'test'}   
    assert compute_shader.extra == {'name': 'test'}
    assert isinstance(compute_shader.ctx, moderngl.Context)

    compute_shader['multiplier'] = 1.0
    compute_shader['multiplier'].value = 1.0
    compute_shader.run()
    assert compute_shader.get('multiplier', None) is not None
    assert compute_shader == compute_shader
    assert [i for i in compute_shader] == ['multiplier', 'something_in', 'something_out']
    compute_shader.release()



import struct
import pytest
import moderngl


@pytest.fixture(scope='module')
def prog(ctx_static):
    return ctx_static.program(
        vertex_shader='''
            #version 330

            in int a_in;
            in int b_in;
            in int c_in;
            in int d_in;

            out int a_out;
            out int b_out;
            out int c_out;
            out int d_out;

            void main() {
                a_out = a_in * 2;
                b_out = b_in * 2;
                c_out = c_in * 2;
                d_out = d_in * 2;
            }
        ''',
        varyings=['a_out', 'b_out', 'c_out', 'd_out']
    )


@pytest.fixture(scope='module', autouse=True)
def fbo(ctx_static):
    """Bind an fbo to avoid GL_INVALID_FRAMEBUFFER_OPERATION with transforms"""
    fbo = ctx_static.simple_framebuffer((2, 2))
    fbo.use()


def test_padding_1(ctx, prog):
    buf = ctx.buffer(struct.pack('=ixi12xii', 1, 2, 3, 4))
    res = ctx.buffer(reserve=16)

    vao = ctx.vertex_array(prog, [
        (buf, 'i x i 12x i i', 'a_in', 'b_in', 'c_in', 'd_in'),
    ])

    vao.transform(res, moderngl.POINTS)
    a, b, c, d = struct.unpack('=iiii', res.read())

    assert ctx.error == 'GL_NO_ERROR'

    assert a == 2
    assert b == 4
    assert c == 6
    assert d == 8


def test_padding_2(ctx, prog):
    buf = ctx.buffer(struct.pack('=i8xi8xi8xi8x', 1, 2, 3, 4))
    res = ctx.buffer(reserve=64)

    vao = ctx.vertex_array(prog, [
        (buf, 'i 8x', 'a_in'),
    ])

    vao.transform(res, moderngl.POINTS)
    assert vao.vertices == 4

    a1, a2, a3, a4 = struct.unpack('=i12xi12xi12xi12x', res.read())

    assert a1 == 2
    assert a2 == 4
    assert a3 == 6
    assert a4 == 8


def test_padding_3(ctx, prog):
    buf = ctx.buffer(struct.pack('=128xiiii', 1, 2, 3, 4))
    res = ctx.buffer(reserve=16)

    vao = ctx.vertex_array(prog, [
        (buf, '128x i i i i', 'a_in', 'b_in', 'c_in', 'd_in'),
    ])

    vao.transform(res, moderngl.POINTS)
    a, b, c, d = struct.unpack('=iiii', res.read())

    assert a == 2
    assert b == 4
    assert c == 6
    assert d == 8




def test_1(ctx):
    buf1 = ctx.buffer(b'abc')
    buf2 = ctx.buffer(reserve=3)
    ctx.copy_buffer(buf2, buf1)
    assert buf2.read() == b'abc'


def test_2(ctx):
    buf1 = ctx.buffer(b'abcxyz123')
    buf2 = ctx.buffer(reserve=12)
    ctx.copy_buffer(buf2, buf1, 3, read_offset=3, write_offset=0)
    ctx.copy_buffer(buf2, buf1, 3, read_offset=0, write_offset=3)
    ctx.copy_buffer(buf2, buf1, 3, read_offset=6, write_offset=6)
    ctx.copy_buffer(buf2, buf1, 3, read_offset=3, write_offset=9)
    assert buf2.read() == b'xyzabc123xyz'



import moderngl
import pytest


def test_renderbuffer(ctx):
    ctx.renderbuffer((64, 64))


def test_multisample_renderbuffer(ctx):
    if ctx.max_samples < 2:
        pytest.skip('multisampling is not supported')

    ctx.renderbuffer((64, 64), samples=2)


def test_depth_renderbuffer(ctx):
    ctx.depth_renderbuffer((64, 64))


def test_multisample_depth_renderbuffer(ctx):
    if ctx.max_samples < 2:
        pytest.skip('multisampling is not supported')

    ctx.depth_renderbuffer((64, 64), samples=2)


def test_renderbuffer_invalid_samples(ctx):
    with pytest.raises(moderngl.Error, match='sample'):
        ctx.renderbuffer((64, 64), samples=3)



from array import array
import numpy as np
import numpy.testing as npt
import moderngl


def test_properties(ctx):
    prog = ctx.program(
        vertex_shader="""
        #version 330
        in vec2 pos;
        in vec2 velocity;
        out vec2 out_pos;
        void main() {
            out_pos = pos + velocity;
        }
        """,
        varyings=["out_pos"],
    )
    buffer = ctx.buffer(array('f', range(16)))
    vao = ctx.vertex_array(prog, [(buffer, '2f', 'pos')])

    assert vao.program == prog
    assert vao.vertices == 8
    assert vao.index_buffer is None
    assert vao.index_element_size == 4
    assert vao.mode == moderngl.POINTS
    assert vao.instances == 1
    assert vao.glo > 0

    vao.mode = moderngl.LINES
    assert vao.mode == moderngl.LINES
    vao.vertices = 2
    assert vao.vertices == 2
    vao.instances = 2
    assert vao.instances == 2


def test_padding(ctx):
    prog = ctx.program(
        vertex_shader="""
        #version 330
        in vec2 pos;
        in vec2 velocity;
        out vec2 out_pos;
        void main() {
            out_pos = pos + velocity;
        }
        """,
        varyings=["out_pos"],
    )
    buffer = ctx.buffer(array('f', range(16)))
    ctx.vertex_array(prog, [(buffer, '2f 2x4', 'pos')])
    ctx.vertex_array(prog, [(buffer, '2f 2f', 'pos', 'velocity')])


def test_empty(ctx):
    prog = ctx.program(
        vertex_shader="""
        #version 330
        in vec2 pos;
        in vec2 velocity;
        out vec2 out_pos;
        void main() {
            out_pos = pos + velocity;
        }
        """,
    )
    ctx.vertex_array(prog, [])


# def test_optional(ctx):
#     prog = ctx.program(
#         vertex_shader="""
#         #version 330
#         in vec2 pos;
#         in vec2 velocity;
#         in vec4 color;
#         out vec2 out_pos;
#         void main() {
#             out_pos = pos + velocity;
#         }
#         """,
#     )
#     buffer = ctx.buffer(array('f', range(16)))
#     ctx.vertex_array(prog, [(buffer, '2f 2f 4f', 'pos', 'velocity', 'color?')])


def test_1(ctx):
    prog = ctx.program(
        vertex_shader='''
            #version 330
            in vec4 in_vert;
            out vec4 out_vert;
            void main() {
                out_vert = in_vert;
            }
        ''',
        varyings=['out_vert']
    )

    vbo1 = ctx.buffer(np.array([4.0, 2.0, 7.5, 1.8], dtype='f4').tobytes())
    vbo2 = ctx.buffer(reserve=vbo1.size)
    vao = ctx.simple_vertex_array(prog, vbo1, 'in_vert')
    assert vao.mode == moderngl.POINTS

    vao.transform(vbo2, moderngl.POINTS)
    res = np.frombuffer(vbo2.read(), dtype='f4')
    npt.assert_almost_equal(res, [4.0, 2.0, 7.5, 1.8])


def test_2(ctx):
    prog = ctx.program(
        vertex_shader='''
            #version 330

            in vec4 in_vert;
            out vec4 out_vert;

            void main() {
                out_vert = in_vert;
            }
        ''',
        varyings=['out_vert']
    )

    vbo1 = ctx.buffer(np.array([4.0, 2.0, 7.5, 1.8], dtype='f4').tobytes())
    vbo2 = ctx.buffer(reserve=16)

    vao = ctx.vertex_array(prog, [(vbo1, '4f', 'in_vert')])
    vao.transform(vbo2, moderngl.POINTS, vertices=1)
    res = np.frombuffer(vbo2.read(), dtype='f4')
    npt.assert_almost_equal(res, [4.0, 2.0, 7.5, 1.8])

    vao = ctx.vertex_array(prog, [(vbo1, '3f', 'in_vert')])
    vao.transform(vbo2, moderngl.POINTS, vertices=1)
    res = np.frombuffer(vbo2.read(), dtype='f4')
    npt.assert_almost_equal(res, [4.0, 2.0, 7.5, 1.0])

    vao = ctx.vertex_array(prog, [(vbo1, '2f', 'in_vert')])
    vao.transform(vbo2, moderngl.POINTS, vertices=1)
    res = np.frombuffer(vbo2.read(), dtype='f4')
    npt.assert_almost_equal(res, [4.0, 2.0, 0.0, 1.0])

    vao = ctx.vertex_array(prog, [(vbo1, '1f', 'in_vert')])
    vao.transform(vbo2, moderngl.POINTS, vertices=1)
    res = np.frombuffer(vbo2.read(), dtype='f4')
    npt.assert_almost_equal(res, [4.0, 0.0, 0.0, 1.0])


def test_3(ctx):
    prog = ctx.program(
        vertex_shader='''
            #version 330

            in mat4 in_mat;
            out mat4 out_mat;

            void main() {
                out_mat = in_mat;
            }
        ''',
        varyings=['out_mat']
    )

    vbo1 = ctx.buffer(np.arange(1, 17, dtype='f4').tobytes())
    vbo2 = ctx.buffer(reserve=64)

    vao = ctx.vertex_array(prog, [(vbo1, '16f', 'in_mat')])
    vao.transform(vbo2, moderngl.POINTS, vertices=1)
    res = np.frombuffer(vbo2.read(), dtype='f4')
    npt.assert_almost_equal(res, [
        1, 2, 3, 4,
        5, 6, 7, 8,
        9, 10, 11, 12,
        13, 14, 15, 16,
    ])

    vao = ctx.vertex_array(prog, [(vbo1, '12f', 'in_mat')])
    vao.transform(vbo2, moderngl.POINTS, vertices=1)
    res = np.frombuffer(vbo2.read(), dtype='f4')
    npt.assert_almost_equal(res, [
        1, 2, 3, 1,
        4, 5, 6, 1,
        7, 8, 9, 1,
        10, 11, 12, 1,
    ])

    vao = ctx.vertex_array(prog, [(vbo1, '8f', 'in_mat')])
    vao.transform(vbo2, moderngl.POINTS, vertices=1)
    res = np.frombuffer(vbo2.read(), dtype='f4')
    npt.assert_almost_equal(res, [
        1, 2, 0, 1,
        3, 4, 0, 1,
        5, 6, 0, 1,
        7, 8, 0, 1,
    ])

    vao = ctx.vertex_array(prog, [(vbo1, '4f', 'in_mat')])
    vao.transform(vbo2, moderngl.POINTS, vertices=1)
    res = np.frombuffer(vbo2.read(), dtype='f4')
    npt.assert_almost_equal(res, [
        1, 0, 0, 1,
        2, 0, 0, 1,
        3, 0, 0, 1,
        4, 0, 0, 1,
    ])


def test_4(ctx):
    prog = ctx.program(
        vertex_shader='''
            #version 330

            in mat2 in_mat;
            out mat2 out_mat;

            void main() {
                out_mat = in_mat;
            }
        ''',
        varyings=['out_mat']
    )

    vbo1 = ctx.buffer(np.array([4.0, 2.0, 7.5, 1.8], dtype='f4').tobytes())
    vbo2 = ctx.buffer(reserve=16)

    vao = ctx.vertex_array(prog, [(vbo1, '4f', 'in_mat')])
    vao.transform(vbo2, moderngl.POINTS, vertices=1)
    res = np.frombuffer(vbo2.read(), dtype='f4')
    npt.assert_almost_equal(res, [4.0, 2.0, 7.5, 1.8])

    vao = ctx.vertex_array(prog, [(vbo1, '2f', 'in_mat')])
    vao.transform(vbo2, moderngl.POINTS, vertices=1)
    res = np.frombuffer(vbo2.read(), dtype='f4')
    npt.assert_almost_equal(res, [4.0, 0.0, 2.0, 0.0])




def test_framebuffer_read_channels(ctx):
    fbo = ctx.framebuffer([
        ctx.renderbuffer((4, 4)),
    ])

    fbo.clear(65 / 255, 66 / 255, 67 / 255, 68 / 255)

    assert fbo.read(components=1) == b'A' * 16
    assert fbo.read(components=2) == b'AB' * 16
    assert fbo.read(components=3) == b'ABC' * 16
    assert fbo.read(components=4) == b'ABCD' * 16
